@@ -1,5 +1,5 @@
 /* Copyright (C) 2002-2005 RealVNC Ltd.  All Rights Reserved.
- * Copyright 2009 Pierre Ossman for Cendio AB
+ * Copyright 2009-2011 Pierre Ossman for Cendio AB
  * 
  * This is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -17,10 +17,23 @@
  * USA.
  */
 
+// Debug output on what the congestion control is up to
+#undef CONGESTION_DEBUG
+
+#include <sys/time.h>
+
+#ifdef CONGESTION_DEBUG
+#include <sys/socket.h>
+#include <netinet/in.h>
+#include <netinet/tcp.h>
+#endif
+
+#include <network/TcpSocket.h>
 #include <rfb/VNCSConnectionST.h>
 #include <rfb/LogWriter.h>
-#include <rfb/secTypes.h>
+#include <rfb/Security.h>
 #include <rfb/screenTypes.h>
+#include <rfb/fenceTypes.h>
 #include <rfb/ServerCore.h>
 #include <rfb/ComparingUpdateTracker.h>
 #include <rfb/KeyRemapper.h>
@@ -32,13 +45,36 @@ using namespace rfb;
 
 static LogWriter vlog("VNCSConnST");
 
+// This window should get us going fairly fast on a decent bandwidth network.
+// If it's too high, it will rapidly be reduced and stay low.
+static const unsigned INITIAL_WINDOW = 16384;
+
+// TCP's minimal window is 3*MSS. But since we don't know the MSS, we
+// make a guess at 4 KiB (it's probaly a bit higher).
+static const unsigned MINIMUM_WINDOW = 4096;
+
+// The current default maximum window for Linux (4 MiB). Should be a good
+// limit for now...
+static const unsigned MAXIMUM_WINDOW = 4194304;
+
+struct RTTInfo {
+  struct timeval tv;
+  int offset;
+  unsigned inFlight;
+};
+
 VNCSConnectionST::VNCSConnectionST(VNCServerST* server_, network::Socket *s,
                                    bool reverse)
-  : SConnection(server_->securityFactory, reverse), sock(s), server(server_),
+  : SConnection(reverse), sock(s), inProcessMessages(false),
+    syncFence(false), fenceFlags(0), fenceDataLen(0), fenceData(NULL),
+    baseRTT(-1), minRTT(-1), seenCongestion(false), pingCounter(0),
+    ackedOffset(0), sentOffset(0), congWindow(0), congestionTimer(this),
+    server(server_),
     updates(false), image_getter(server->useEconomicTranslate),
     drawRenderedCursor(false), removeRenderedCursor(false),
-    pointerEventTime(0), accessRights(AccessDefault),
-    startTime(time(0))
+    continuousUpdates(false),
+    updateTimer(this), pointerEventTime(0),
+    accessRights(AccessDefault), startTime(time(0))
 {
   setStreams(&sock->inStream(), &sock->outStream());
   peerEndpoint.buf = sock->getPeerEndpoint();
@@ -69,6 +105,7 @@ VNCSConnectionST::~VNCSConnectionST()
   // Remove this client from the server
   server->clients.remove(this);
 
+  delete [] fenceData;
 }
 
 
@@ -111,22 +148,30 @@ void VNCSConnectionST::processMessages()
   try {
     // - Now set appropriate socket timeouts and process data
     setSocketTimeouts();
-    bool clientsReadyBefore = server->clientsReadyForUpdate();
+
+    inProcessMessages = true;
+
+    // Get the underlying TCP layer to build large packets if we send
+    // multiple small responses.
+    network::TcpSocket::cork(sock->getFd(), true);
 
     while (getInStream()->checkNoWait(1)) {
       processMsg();
+      if (syncFence) {
+        writer()->writeFence(fenceFlags, fenceDataLen, fenceData);
+        syncFence = false;
+      }
     }
 
-    // If there were update requests, try to send a framebuffer update.
-    // We don't send updates immediately on requests as this way, we
-    // give higher priority to user actions such as keyboard and
-    // pointer events.
-    if (!requested.is_empty()) {
-      writeFramebufferUpdate();
-    }
+    // Flush out everything in case we go idle after this.
+    network::TcpSocket::cork(sock->getFd(), false);
 
-    if (!clientsReadyBefore && !requested.is_empty())
-      server->desktop->framebufferUpdateRequest();
+    inProcessMessages = false;
+
+    // If there were anything requiring an update, try to send it here.
+    // We wait until now with this to aggregate responses and to give 
+    // higher priority to user actions such as keyboard and pointer events.
+    writeFramebufferUpdate();
   } catch (rdr::EndOfStream&) {
     close("Clean disconnection");
   } catch (rdr::Exception &e) {
@@ -134,15 +179,6 @@ void VNCSConnectionST::processMessages()
   }
 }
 
-void VNCSConnectionST::writeFramebufferUpdateOrClose()
-{
-  try {
-    writeFramebufferUpdate();
-  } catch(rdr::Exception &e) {
-    close(e.str());
-  }
-}
-
 void VNCSConnectionST::pixelBufferChange()
 {
   try {
@@ -186,27 +222,25 @@ void VNCSConnectionST::pixelBufferChange()
     updates.add_changed(server->pb->getRect());
     vlog.debug("pixel buffer changed - re-initialising image getter");
     image_getter.init(server->pb, cp.pf(), writer());
-    if (writer()->needFakeUpdate())
-      writeFramebufferUpdate();
+    writeFramebufferUpdate();
   } catch(rdr::Exception &e) {
     close(e.str());
   }
 }
 
-void VNCSConnectionST::screenLayoutChange(rdr::U16 reason)
+void VNCSConnectionST::writeFramebufferUpdateOrClose()
 {
   try {
-    if (!authenticated())
-      return;
+    writeFramebufferUpdate();
+  } catch(rdr::Exception &e) {
+    close(e.str());
+  }
+}
 
-    cp.screenLayout = server->screenLayout;
-    if (state() == RFBSTATE_NORMAL) {
-      writer()->writeExtendedDesktopSize(reason, 0, cp.width, cp.height,
-                                         cp.screenLayout);
-    }
-
-    if (writer()->needFakeUpdate())
-      writeFramebufferUpdate();
+void VNCSConnectionST::screenLayoutChangeOrClose(rdr::U16 reason)
+{
+  try {
+    screenLayoutChange(reason);
   } catch(rdr::Exception &e) {
     close(e.str());
   }
@@ -221,7 +255,7 @@ void VNCSConnectionST::setColourMapEntriesOrClose(int firstColour,int nColours)
   }
 }
 
-void VNCSConnectionST::bell()
+void VNCSConnectionST::bellOrClose()
 {
   try {
     if (state() == RFBSTATE_NORMAL) writer()->writeBell();
@@ -230,7 +264,7 @@ void VNCSConnectionST::bell()
   }
 }
 
-void VNCSConnectionST::serverCutText(const char *str, int len)
+void VNCSConnectionST::serverCutTextOrClose(const char *str, int len)
 {
   try {
     if (!(accessRights & AccessCutText)) return;
@@ -243,15 +277,10 @@ void VNCSConnectionST::serverCutText(const char *str, int len)
 }
 
 
-void VNCSConnectionST::setDesktopName(const char *name)
+void VNCSConnectionST::setDesktopNameOrClose(const char *name)
 {
-  cp.setName(name);
   try {
-    if (state() == RFBSTATE_NORMAL) {
-      if (!writer()->writeSetDesktopName()) {
-	fprintf(stderr, "Client does not support desktop rename\n");
-      }
-    }
+    setDesktopName(name);
   } catch(rdr::Exception& e) {
     close(e.str());
   }
@@ -297,6 +326,16 @@ int VNCSConnectionST::checkIdleTimeout()
   return secsToMillis(timeLeft);
 }
 
+
+bool VNCSConnectionST::getComparerState()
+{
+  // We interpret a low compression level as an indication that the client
+  // wants to prioritise CPU usage over bandwidth, and hence disable the
+  // comparing update tracker.
+  return (cp.compressLevel == -1) || (cp.compressLevel > 1);
+}
+
+
 // renderedCursorChange() is called whenever the server-side rendered cursor
 // changes shape or position.  It ensures that the next update will clean up
 // the old rendered cursor and if necessary draw the new rendered cursor.
@@ -304,9 +343,12 @@ int VNCSConnectionST::checkIdleTimeout()
 void VNCSConnectionST::renderedCursorChange()
 {
   if (state() != RFBSTATE_NORMAL) return;
-  removeRenderedCursor = true;
-  if (needRenderedCursor())
+  if (!renderedCursorRect.is_empty())
+    removeRenderedCursor = true;
+  if (needRenderedCursor()) {
     drawRenderedCursor = true;
+    writeFramebufferUpdateOrClose();
+  }
 }
 
 // needRenderedCursor() returns true if this client needs the server-side
@@ -321,10 +363,9 @@ void VNCSConnectionST::renderedCursorChange()
 
 bool VNCSConnectionST::needRenderedCursor()
 {
+  bool pointerpos = (!server->cursorPos.equals(pointerEventPos) && (time(0) - pointerEventTime) > 0);
   return (state() == RFBSTATE_NORMAL
-          && (!cp.supportsLocalCursor && !cp.supportsLocalXCursor
-              || (!server->cursorPos.equals(pointerEventPos) &&
-                  (time(0) - pointerEventTime) > 0)));
+          && ((!cp.supportsLocalCursor && !cp.supportsLocalXCursor) || pointerpos));
 }
 
 
@@ -364,6 +405,10 @@ void VNCSConnectionST::authSuccess()
   // - Mark the entire display as "dirty"
   updates.add_changed(server->pb->getRect());
   startTime = time(0);
+
+  // - Bootstrap the congestion control
+  ackedOffset = sock->outStream().length();
+  congWindow = INITIAL_WINDOW;
 }
 
 void VNCSConnectionST::queryConnection(const char* userName)
@@ -521,7 +566,8 @@ void VNCSConnectionST::framebufferUpdateRequest(const Rect& r,bool incremental)
   // Just update the requested region.
   // Framebuffer update will be sent a bit later, see processMessages().
   Region reqRgn(r);
-  requested.assign_union(reqRgn);
+  if (!incremental || !continuousUpdates)
+    requested.assign_union(reqRgn);
 
   if (!incremental) {
     // Non-incremental update - treat as if area requested has changed
@@ -548,8 +594,7 @@ void VNCSConnectionST::setDesktopSize(int fb_width, int fb_height,
   if (!layout.validate(fb_width, fb_height)) {
     writer()->writeExtendedDesktopSize(reasonClient, resultInvalid,
                                        fb_width, fb_height, layout);
-    if (writer()->needFakeUpdate())
-      writeFramebufferUpdate();
+    writeFramebufferUpdate();
     return;
   }
 
@@ -558,18 +603,19 @@ void VNCSConnectionST::setDesktopSize(int fb_width, int fb_height,
   // protocol-wise, but unnecessary.
   result = server->desktop->setScreenLayout(fb_width, fb_height, layout);
 
-  // Always send back a reply to the requesting client
   writer()->writeExtendedDesktopSize(reasonClient, result,
                                      fb_width, fb_height, layout);
-  if (writer()->needFakeUpdate())
-    writeFramebufferUpdate();
 
-  // But only notify other clients on success
+  // Only notify other clients on success
   if (result == resultSuccess) {
     if (server->screenLayout != layout)
         throw Exception("Desktop configured a different screen layout than requested");
     server->notifyScreenLayoutChange(this);
   }
+
+  // but always send back a reply to the requesting client
+  // (do this last as it might throw an exception on socket errors)
+  writeFramebufferUpdate();
 }
 
 void VNCSConnectionST::setInitialColourMap()
@@ -577,6 +623,69 @@ void VNCSConnectionST::setInitialColourMap()
   setColourMapEntries(0, 0);
 }
 
+void VNCSConnectionST::fence(rdr::U32 flags, unsigned len, const char data[])
+{
+  if (flags & fenceFlagRequest) {
+    if (flags & fenceFlagSyncNext) {
+      if (syncFence)
+        vlog.error("Fence trying to synchronise another fence");
+
+      syncFence = true;
+
+      fenceFlags = flags & (fenceFlagBlockBefore | fenceFlagBlockAfter | fenceFlagSyncNext);
+      fenceDataLen = len;
+      delete [] fenceData;
+      if (len > 0) {
+        fenceData = new char[len];
+        memcpy(fenceData, data, len);
+      }
+
+      return;
+    }
+
+    // We handle everything synchronously so we trivially honor these modes
+    flags = flags & (fenceFlagBlockBefore | fenceFlagBlockAfter);
+
+    writer()->writeFence(flags, len, data);
+    return;
+  }
+
+  struct RTTInfo rttInfo;
+
+  switch (len) {
+  case 0:
+    // Initial dummy fence;
+    break;
+  case sizeof(struct RTTInfo):
+    memcpy(&rttInfo, data, sizeof(struct RTTInfo));
+    handleRTTPong(rttInfo);
+    break;
+  default:
+    vlog.error("Fence response of unexpected size received");
+  }
+}
+
+void VNCSConnectionST::enableContinuousUpdates(bool enable,
+                                               int x, int y, int w, int h)
+{
+  Rect rect;
+
+  if (!cp.supportsFence || !cp.supportsContinuousUpdates)
+    throw Exception("Client tried to enable continuous updates when not allowed");
+
+  continuousUpdates = enable;
+
+  rect.setXYWH(x, y, w, h);
+  cuRegion.reset(rect);
+
+  if (enable) {
+    requested.clear();
+    writeFramebufferUpdate();
+  } else {
+    writer()->writeEndOfContinuousUpdates();
+  }
+}
+
 // supportsLocalCursor() is called whenever the status of
 // cp.supportsLocalCursor has changed.  If the client does now support local
 // cursor, we make sure that the old server-side rendered cursor is cleaned up
@@ -585,12 +694,28 @@ void VNCSConnectionST::setInitialColourMap()
 void VNCSConnectionST::supportsLocalCursor()
 {
   if (cp.supportsLocalCursor || cp.supportsLocalXCursor) {
-    removeRenderedCursor = true;
+    if (!renderedCursorRect.is_empty())
+      removeRenderedCursor = true;
     drawRenderedCursor = false;
     setCursor();
   }
 }
 
+void VNCSConnectionST::supportsFence()
+{
+  writer()->writeFence(fenceFlagRequest, 0, NULL);
+}
+
+void VNCSConnectionST::supportsContinuousUpdates()
+{
+  // We refuse to use continuous updates if we cannot monitor the buffer
+  // usage using fences.
+  if (!cp.supportsFence)
+    return;
+
+  writer()->writeEndOfContinuousUpdates();
+}
+
 void VNCSConnectionST::writeSetCursorCallback()
 {
   if (cp.supportsLocalXCursor) {
@@ -626,30 +751,262 @@ void VNCSConnectionST::writeSetCursorCallback()
 }
 
 
+bool VNCSConnectionST::handleTimeout(Timer* t)
+{
+  try {
+    if (t == &updateTimer)
+      writeFramebufferUpdate();
+    else if (t == &congestionTimer)
+      updateCongestion();
+  } catch (rdr::Exception& e) {
+    close(e.str());
+  }
+
+  return false;
+}
+
+
+void VNCSConnectionST::writeRTTPing()
+{
+  struct RTTInfo rttInfo;
+
+  if (!cp.supportsFence)
+    return;
+
+  memset(&rttInfo, 0, sizeof(struct RTTInfo));
+
+  gettimeofday(&rttInfo.tv, NULL);
+  rttInfo.offset = sock->outStream().length();
+  rttInfo.inFlight = rttInfo.offset - ackedOffset;
+
+  // We need to make sure any old update are already processed by the
+  // time we get the response back. This allows us to reliably throttle
+  // back on client overload, as well as network overload.
+  writer()->writeFence(fenceFlagRequest | fenceFlagBlockBefore,
+                       sizeof(struct RTTInfo), (const char*)&rttInfo);
+
+  pingCounter++;
+
+  sentOffset = rttInfo.offset;
+
+  // Let some data flow before we adjust the settings
+  if (!congestionTimer.isStarted())
+    congestionTimer.start(__rfbmin(baseRTT * 2, 100));
+}
+
+void VNCSConnectionST::handleRTTPong(const struct RTTInfo &rttInfo)
+{
+  unsigned rtt, delay;
+  int bdp;
+
+  pingCounter--;
+
+  rtt = msSince(&rttInfo.tv);
+  if (rtt < 1)
+    rtt = 1;
+
+  ackedOffset = rttInfo.offset;
+
+  // Try to estimate wire latency by tracking lowest seen latency
+  if (rtt < baseRTT)
+    baseRTT = rtt;
+
+  if (rttInfo.inFlight > congWindow) {
+    seenCongestion = true;
+
+    // Estimate added delay because of overtaxed buffers
+    delay = (rttInfo.inFlight - congWindow) * baseRTT / congWindow;
+
+    if (delay < rtt)
+      rtt -= delay;
+    else
+      rtt = 1;
+
+    // If we underestimate the congestion window, then we'll get a latency
+    // that's less than the wire latency, which will confuse other portions
+    // of the code.
+    if (rtt < baseRTT)
+      rtt = baseRTT;
+  }
+
+  // We only keep track of the minimum latency seen (for a given interval)
+  // on the basis that we want to avoid continous buffer issue, but don't
+  // mind (or even approve of) bursts.
+  if (rtt < minRTT)
+    minRTT = rtt;
+}
+
+bool VNCSConnectionST::isCongested()
+{
+  int offset;
+
+  // Stuff still waiting in the send buffer?
+  if (sock->outStream().bufferUsage() > 0)
+    return true;
+
+  if (!cp.supportsFence)
+    return false;
+
+  // Idle for too long? (and no data on the wire)
+  //
+  // FIXME: This should really just be one baseRTT, but we're getting
+  //        problems with triggering the idle timeout on each update.
+  //        Maybe we need to use a moving average for the wire latency
+  //        instead of baseRTT.
+  if ((sentOffset == ackedOffset) &&
+      (sock->outStream().getIdleTime() > 2 * baseRTT)) {
+
+#ifdef CONGESTION_DEBUG
+    if (congWindow > INITIAL_WINDOW)
+      fprintf(stderr, "Reverting to initial window (%d KiB) after %d ms\n",
+              INITIAL_WINDOW / 1024, sock->outStream().getIdleTime());
+#endif
+
+    // Close congestion window and allow a transfer
+    // FIXME: Reset baseRTT like Linux Vegas?
+    congWindow = __rfbmin(INITIAL_WINDOW, congWindow);
+
+    return false;
+  }
+
+  offset = sock->outStream().length();
+
+  // FIXME: Should we compensate for non-update data?
+  //        (i.e. use sentOffset instead of offset)
+  if ((offset - ackedOffset) < congWindow)
+    return false;
+
+  // If we just have one outstanding "ping", that means the client has
+  // started receiving our update. In order to not regress compared to
+  // before we had congestion avoidance, we allow another update here.
+  // This could further clog up the tubes, but congestion control isn't
+  // really working properly right now anyway as the wire would otherwise
+  // be idle for at least RTT/2.
+  if (pingCounter == 1)
+    return false;
+
+  return true;
+}
+
+
+void VNCSConnectionST::updateCongestion()
+{
+  unsigned diff;
+
+  if (!seenCongestion)
+    return;
+
+  diff = minRTT - baseRTT;
+
+  if (diff > __rfbmin(100, baseRTT)) {
+    // Way too fast
+    congWindow = congWindow * baseRTT / minRTT;
+  } else if (diff > __rfbmin(50, baseRTT/2)) {
+    // Slightly too fast
+    congWindow -= 4096;
+  } else if (diff < 5) {
+    // Way too slow
+    congWindow += 8192;
+  } else if (diff < 25) {
+    // Too slow
+    congWindow += 4096;
+  }
+
+  if (congWindow < MINIMUM_WINDOW)
+    congWindow = MINIMUM_WINDOW;
+  if (congWindow > MAXIMUM_WINDOW)
+    congWindow = MAXIMUM_WINDOW;
+
+#ifdef CONGESTION_DEBUG
+  fprintf(stderr, "RTT: %d ms (%d ms), Window: %d KiB, Bandwidth: %g Mbps\n",
+          minRTT, baseRTT, congWindow / 1024,
+          congWindow * 8.0 / baseRTT / 1000.0);
+
+#ifdef TCP_INFO
+  struct tcp_info tcp_info;
+  socklen_t tcp_info_length;
+
+  tcp_info_length = sizeof(tcp_info);
+  if (getsockopt(sock->getFd(), SOL_TCP, TCP_INFO,
+                 (void *)&tcp_info, &tcp_info_length) == 0) {
+    fprintf(stderr, "Socket: RTT: %d ms (+/- %d ms) Window %d KiB\n",
+            tcp_info.tcpi_rtt / 1000, tcp_info.tcpi_rttvar / 1000,
+            tcp_info.tcpi_snd_mss * tcp_info.tcpi_snd_cwnd / 1024);
+  }
+#endif
+
+#endif
+
+  minRTT = -1;
+  seenCongestion = false;
+}
+
+
 void VNCSConnectionST::writeFramebufferUpdate()
 {
-  if (state() != RFBSTATE_NORMAL || requested.is_empty())
+  Region req;
+  UpdateInfo ui;
+  bool needNewUpdateInfo;
+
+  updateTimer.stop();
+
+  // We're in the middle of processing a command that's supposed to be
+  // synchronised. Allowing an update to slip out right now might violate
+  // that synchronisation.
+  if (syncFence)
     return;
 
+  // We try to aggregate responses, so don't send out anything whilst we
+  // still have incoming messages. processMessages() will give us another
+  // chance to run once things are idle.
+  if (inProcessMessages)
+    return;
+
+  if (state() != RFBSTATE_NORMAL)
+    return;
+  if (requested.is_empty() && !continuousUpdates)
+    return;
+
+  // Check that we actually have some space on the link and retry in a
+  // bit if things are congested.
+  if (isCongested()) {
+    updateTimer.start(50);
+    return;
+  }
+
+  // In continuous mode, we will be outputting at least three distinct
+  // messages. We need to aggregate these in order to not clog up TCP's
+  // congestion window.
+  network::TcpSocket::cork(sock->getFd(), true);
+
   // First take care of any updates that cannot contain framebuffer data
   // changes.
   if (writer()->needNoDataUpdate()) {
     writer()->writeNoDataUpdate();
     requested.clear();
-    return;
+    if (!continuousUpdates)
+      goto out;
   }
 
   updates.enable_copyrect(cp.useCopyRect);
 
-  server->checkUpdate();
+  // Fetch updates from server object, and see if we are allowed to send
+  // anything right now (the framebuffer might have changed in ways we
+  // haven't yet been informed of).
+  if (!server->checkUpdate())
+    goto out;
 
   // Get the lists of updates. Prior to exporting the data to the `ui' object,
   // getUpdateInfo() will normalize the `updates' object such way that its
   // `changed' and `copied' regions would not intersect.
 
-  UpdateInfo ui;
-  updates.getUpdateInfo(&ui, requested);
-  bool needNewUpdateInfo = false;
+  if (continuousUpdates)
+    req = cuRegion.union_(requested);
+  else
+    req = requested;
+
+  updates.getUpdateInfo(&ui, req);
+  needNewUpdateInfo = false;
 
   // If the previous position of the rendered cursor overlaps the source of the
   // copy, then when the copy happens the corresponding rectangle in the
@@ -677,12 +1034,12 @@ void VNCSConnectionST::writeFramebufferUpdate()
   // Return if there is nothing to send the client.
 
   if (updates.is_empty() && !writer()->needFakeUpdate() && !drawRenderedCursor)
-    return;
+    goto out;
 
   // The `updates' object could change, make sure we have valid update info.
 
   if (needNewUpdateInfo)
-    updates.getUpdateInfo(&ui, requested);
+    updates.getUpdateInfo(&ui, req);
 
   // If the client needs a server-side rendered cursor, work out the cursor
   // rectangle.  If it's empty then don't bother drawing it, but if it overlaps
@@ -692,7 +1049,7 @@ void VNCSConnectionST::writeFramebufferUpdate()
   if (needRenderedCursor()) {
     renderedCursorRect
       = (server->renderedCursor.getRect(server->renderedCursorTL)
-         .intersect(requested.get_bounding_rect()));
+         .intersect(req.get_bounding_rect()));
 
     if (renderedCursorRect.is_empty()) {
       drawRenderedCursor = false;
@@ -708,7 +1065,7 @@ void VNCSConnectionST::writeFramebufferUpdate()
 
     //if (drawRenderedCursor) {
     //  updates.subtract(renderedCursorRect);
-    //  updates.getUpdateInfo(&ui, requested);
+    //  updates.getUpdateInfo(&ui, req);
     //}
   }
 
@@ -723,19 +1080,36 @@ void VNCSConnectionST::writeFramebufferUpdate()
     std::vector<Rect>::const_iterator i;
     ui.changed.get_rects(&rects);
     for (i = rects.begin(); i != rects.end(); i++) {
-      if (i->width() && i->height())
-        nRects += writer()->getNumRects(*i);
+      if (i->width() && i->height()) {
+        int nUpdateRects = writer()->getNumRects(*i);
+        if (nUpdateRects == 0 && cp.currentEncoding() == encodingTight) {
+          nRects = 0xFFFF;  break;
+        }
+        else
+          nRects += nUpdateRects;
+      }
     }
+
+    writeRTTPing();
     
     writer()->writeFramebufferUpdateStart(nRects);
+
     Region updatedRegion;
     writer()->writeRects(ui, &image_getter, &updatedRegion);
     updates.subtract(updatedRegion);
+
     if (drawRenderedCursor)
       writeRenderedCursorRect();
+
     writer()->writeFramebufferUpdateEnd();
+
+    writeRTTPing();
+
     requested.clear();
   }
+
+out:
+  network::TcpSocket::cork(sock->getFd(), false);
 }
 
 
@@ -756,15 +1130,33 @@ void VNCSConnectionST::writeRenderedCursorRect()
   drawRenderedCursor = false;
 }
 
+void VNCSConnectionST::screenLayoutChange(rdr::U16 reason)
+{
+  if (!authenticated())
+    return;
+
+  cp.screenLayout = server->screenLayout;
+
+  if (state() != RFBSTATE_NORMAL)
+    return;
+
+  writer()->writeExtendedDesktopSize(reason, 0, cp.width, cp.height,
+                                     cp.screenLayout);
+  writeFramebufferUpdate();
+}
+
 void VNCSConnectionST::setColourMapEntries(int firstColour, int nColours)
 {
-  if (!readyForSetColourMapEntries) return;
-  if (server->pb->getPF().trueColour) return;
+  if (!readyForSetColourMapEntries)
+    return;
+  if (server->pb->getPF().trueColour)
+    return;
 
-  image_getter.setColourMapEntries(firstColour, nColours, writer());
+  image_getter.setColourMapEntries(firstColour, nColours);
 
   if (cp.pf().trueColour) {
     updates.add_changed(server->pb->getRect());
+    writeFramebufferUpdate();
   }
 }
 
@@ -775,10 +1167,28 @@ void VNCSConnectionST::setColourMapEntries(int firstColour, int nColours)
 
 void VNCSConnectionST::setCursor()
 {
-  if (state() != RFBSTATE_NORMAL || !cp.supportsLocalCursor) return;
+  if (state() != RFBSTATE_NORMAL)
+    return;
+  if (!cp.supportsLocalCursor)
+    return;
+
   writer()->cursorChange(this);
-  if (writer()->needFakeUpdate())
-    writeFramebufferUpdate();
+  writeFramebufferUpdate();
+}
+
+void VNCSConnectionST::setDesktopName(const char *name)
+{
+  cp.setName(name);
+
+  if (state() != RFBSTATE_NORMAL)
+    return;
+
+  if (!writer()->writeSetDesktopName()) {
+    fprintf(stderr, "Client does not support desktop rename\n");
+    return;
+  }
+
+  writeFramebufferUpdate();
 }
 
 void VNCSConnectionST::setSocketTimeouts()
@@ -805,10 +1215,10 @@ void VNCSConnectionST::setStatus(int status)
     accessRights = accessRights | AccessPtrEvents | AccessKeyEvents | AccessView;
     break;
   case 1:
-    accessRights = accessRights & !(AccessPtrEvents | AccessKeyEvents) | AccessView;
+    accessRights = accessRights & ~(AccessPtrEvents | AccessKeyEvents) | AccessView;
     break;
   case 2:
-    accessRights = accessRights & !(AccessPtrEvents | AccessKeyEvents | AccessView);
+    accessRights = accessRights & ~(AccessPtrEvents | AccessKeyEvents | AccessView);
     break;
   }
   framebufferUpdateRequest(server->pb->getRect(), false);
