@@ -30,6 +30,13 @@ namespace rfb {
 
   class VNCServer : public UpdateTracker {
   public:
+    // blockUpdates()/unblockUpdates() tells the server that the pixel buffer
+    // is currently in flux and may not be accessed. The attributes of the
+    // pixel buffer may still be accessed, but not the frame buffer itself.
+    // Note that access must be unblocked the exact same number of times it
+    // was blocked.
+    virtual void blockUpdates() = 0;
+    virtual void unblockUpdates() = 0;
 
     // setPixelBuffer() tells the server to use the given pixel buffer (and
     // optionally a modified screen layout).  If this differs in size from
@@ -58,18 +65,10 @@ namespace rfb {
     // bell() tells the server that it should make all clients make a bell sound.
     virtual void bell() = 0;
 
-    // clientsReadyForUpdate() returns true if there is at least one client
-    // waiting for an update, false if no clients are ready.
-    virtual bool clientsReadyForUpdate() = 0;
-
     // - Close all currently-connected clients, by calling
     //   their close() method with the supplied reason.
     virtual void closeClients(const char* reason) = 0;
 
-    // tryUpdate() causes the server to attempt to send updates to any waiting
-    // clients.
-    virtual void tryUpdate() = 0;
-
     // setCursor() tells the server that the cursor has changed.  The
     // cursorData argument contains width*height pixel values in the pixel
     // buffer's format.  The mask argument is a bitmask with a 1-bit meaning
