@@ -1,6 +1,6 @@
 /* Copyright (c) 1993  X Consortium
    Copyright (C) 2002-2005 RealVNC Ltd.  All Rights Reserved.
-   Copyright (C) 2009 Pierre Ossman for Cendio AB
+   Copyright 2009 Pierre Ossman for Cendio AB
 
 Permission is hereby granted, free of charge, to any person obtaining
 a copy of this software and associated documentation files (the
@@ -37,6 +37,7 @@ from the X Consortium.
 #include <rfb/LogWriter.h>
 #include <network/TcpSocket.h>
 #include "vncExtInit.h"
+#include "xorg-version.h"
 
 extern "C" {
 #define class c_class
@@ -80,14 +81,23 @@ extern "C" {
 #endif /* RANDR */
 #include <X11/keysym.h>
   extern char buildtime[];
+#if XORG >= 17
+#undef VENDOR_RELEASE
+#undef VENDOR_STRING
+#include "version-config.h"
+#include "site.h"
+#endif
 #undef class
 #undef public
 }
 
-#define XVNCVERSION "TigerVNC 1.0.0"
-#define XVNCCOPYRIGHT ("Copyright (C) 2002-2005 RealVNC Ltd.\n" \
-		       "Copyright (C) 2000-2006 Constantin Kaplinsky\n" \
-		       "Copyright (C) 2004-2009 Peter Astrand for Cendio AB\n" \
+#if XORG >= 110
+#define Xalloc malloc
+#define Xfree free
+#endif
+
+#define XVNCVERSION "TigerVNC 1.2.0"
+#define XVNCCOPYRIGHT ("Copyright (C) 1999-2011 TigerVNC Team and many others (see README.txt)\n" \
                        "See http://www.tigervnc.org for information on TigerVNC.\n")
 
 
@@ -154,9 +164,10 @@ static int lastScreen = -1;
 static Bool Render = TRUE;
 
 static bool displaySpecified = false;
-static bool wellKnownSocketsCreated = false;
 static char displayNumStr[16];
 
+char *listenaddr = NULL;
+
 
 static void
 vfbInitializePixmapDepths(void)
@@ -200,6 +211,24 @@ static void vfbFreeFramebufferMemory(vfbFramebufferInfoPtr pfb);
 
 extern "C" {
 
+#ifdef DPMSExtension
+    /* Why support DPMS? Because stupid modern desktop environments
+       such as Unity 2D on Ubuntu 11.10 crashes if DPMS is not
+       available. (DPMSSet is called by dpms.c, but the return value
+       is ignored.) */
+int DPMSSet(ClientPtr client, int level)
+{
+    return Success;
+}
+
+Bool DPMSSupported()
+{
+    /* Causes DPMSCapable to return false, meaning no devices are DPMS
+       capable */
+    return FALSE;
+}
+#endif
+
 void ddxGiveUp()
 {
     int i;
@@ -277,6 +306,8 @@ ddxUseMsg()
     ErrorF("-depth D               set screen 0's depth\n");
     ErrorF("-pixelformat fmt       set pixel format (rgbNNN or bgrNNN)\n");
     ErrorF("-inetd                 has been launched from inetd\n");
+    ErrorF("-interface IP_address  listen on specified interface\n");
+    ErrorF("-noclipboard           disable clipboard settings modification via vncconfig utility\n");
     ErrorF("\nVNC parameters:\n");
 
     fprintf(stderr,"\n"
@@ -299,7 +330,7 @@ static
 bool displayNumFree(int num)
 {
     try {
-	network::TcpListener l(6000+num);
+	network::TcpListener l(NULL, 6000+num);
     } catch (rdr::Exception& e) {
 	return false;
     }
@@ -325,6 +356,7 @@ ddxProcessArgument(int argc, char *argv[], int i)
 	firstTime = FALSE;
 	rfb::initStdIOLoggers();
 	rfb::LogWriter::setLogParams("*:stderr:30");
+	rfb::Configuration::enableServerParams();
     }
 
     if (argv[i][0] ==  ':')
@@ -537,6 +569,28 @@ ddxProcessArgument(int argc, char *argv[], int i)
 	
 	return 1;
     }
+
+    if (strcmp(argv[i], "-interface") == 0 ||
+	strcmp(argv[i], "-i") == 0) {
+	if (++i >= argc) {
+	    UseMsg();
+	    return 2;
+	}
+
+	if (listenaddr != NULL) /* Only first -interface is valid */
+		return 2;
+
+	listenaddr = strdup(argv[i]);
+	if (listenaddr == NULL)
+	    FatalError("Not enough memory");
+
+	return 2;
+    }
+
+    if (strcmp(argv[i], "-noclipboard") == 0) {
+	noclipboard = true;
+	return 1;
+    }
     
     if (rfb::Configuration::setParam(argv[i]))
 	return 1;
@@ -602,7 +656,11 @@ vfbInstallColormap(ColormapPtr pmap)
 
 	for (i = 0; i < entries; i++)  ppix[i] = i;
 	/* XXX truecolor */
+#if XORG < 19
 	QueryColors(pmap, entries, ppix, prgb);
+#else
+	QueryColors(pmap, entries, ppix, prgb, serverClient);
+#endif
 
 	for (i = 0; i < entries; i++) { /* convert xrgbs to xColorItems */
 	    defs[i].pixel = ppix[i] & 0xff; /* change pixel to index */
@@ -736,7 +794,7 @@ vfbCrossScreen (ScreenPtr pScreen, Bool entering)
 }
 
 static Bool vfbRealizeCursor(
-#ifdef XORG_16
+#if XORG >= 16
 			     DeviceIntPtr pDev,
 #endif
 			     ScreenPtr pScreen, CursorPtr pCursor) {
@@ -744,7 +802,7 @@ static Bool vfbRealizeCursor(
 }
 
 static Bool vfbUnrealizeCursor(
-#ifdef XORG_16
+#if XORG >= 16
 			       DeviceIntPtr pDev,
 #endif
 			       ScreenPtr pScreen, CursorPtr pCursor) {
@@ -752,7 +810,7 @@ static Bool vfbUnrealizeCursor(
 }
 
 static void vfbSetCursor(
-#ifdef XORG_16
+#if XORG >= 16
 			 DeviceIntPtr pDev,
 #endif
 			 ScreenPtr pScreen, CursorPtr pCursor, int x, int y) 
@@ -760,14 +818,14 @@ static void vfbSetCursor(
 }
 
 static void vfbMoveCursor(
-#ifdef XORG_16
+#if XORG >= 16
 			  DeviceIntPtr pDev,
 #endif
 			  ScreenPtr pScreen, int x, int y) 
 {
 }
 
-#ifdef XORG_16
+#if XORG >= 16
 static Bool
 vfbDeviceCursorInitialize(DeviceIntPtr pDev, ScreenPtr pScreen)
 {   
@@ -785,7 +843,7 @@ static miPointerSpriteFuncRec vfbPointerSpriteFuncs = {
     vfbUnrealizeCursor,
     vfbSetCursor,
     vfbMoveCursor
-#ifdef XORG_16
+#if XORG >= 16
     , vfbDeviceCursorInitialize,
     vfbDeviceCursorCleanup
 #endif
@@ -801,7 +859,6 @@ static miPointerScreenFuncRec vfbPointerCursorFuncs = {
 
 static Bool vncRandRGetInfo (ScreenPtr pScreen, Rotation *rotations)
 {
-  vfbScreenInfoPtr pvfb = &vfbScreens[pScreen->myNum];
   Bool ret, gotCurrent = FALSE;
   int i;
 
@@ -850,13 +907,19 @@ static Bool vncRandRGetInfo (ScreenPtr pScreen, Rotation *rotations)
 static void
 xf86SetRootClip (ScreenPtr pScreen, Bool enable)
 {
+#if XORG < 19
     WindowPtr	pWin = WindowTable[pScreen->myNum];
+#else
+    WindowPtr	pWin = pScreen->root;
+#endif
     WindowPtr	pChild;
     Bool	WasViewable = (Bool)(pWin->viewable);
     Bool	anyMarked = FALSE;
+#if XORG < 110
     RegionPtr	pOldClip = NULL, bsExposed;
 #ifdef DO_SAVE_UNDERS
     Bool	dosave = FALSE;
+#endif
 #endif
     WindowPtr   pLayerWin;
     BoxRec	box;
@@ -915,11 +978,13 @@ xf86SetRootClip (ScreenPtr pScreen, Bool enable)
     
     if (WasViewable)
     {
+#if XORG < 110
 	if (pWin->backStorage)
 	{
 	    pOldClip = REGION_CREATE(pScreen, NullBox, 1);
 	    REGION_COPY(pScreen, pOldClip, &pWin->clipList);
 	}
+#endif
 
 	if (pWin->firstChild)
 	{
@@ -933,7 +998,7 @@ xf86SetRootClip (ScreenPtr pScreen, Bool enable)
 	    anyMarked = TRUE;
 	}
 
-#ifdef DO_SAVE_UNDERS
+#if XORG < 110 && defined(DO_SAVE_UNDERS)
 	if (DO_SAVE_UNDERS(pWin))
 	{
 	    dosave = (*pScreen->ChangeSaveUnder)(pLayerWin, pLayerWin);
@@ -944,6 +1009,7 @@ xf86SetRootClip (ScreenPtr pScreen, Bool enable)
 	    (*pScreen->ValidateTree)(pWin, NullWindow, VTOther);
     }
 
+#if XORG < 110
     if (pWin->backStorage &&
 	((pWin->backingStore == Always) || WasViewable))
     {
@@ -966,11 +1032,13 @@ xf86SetRootClip (ScreenPtr pScreen, Bool enable)
 	    REGION_DESTROY(pScreen, bsExposed);
 	}
     }
+#endif
     if (WasViewable)
     {
 	if (anyMarked)
 	    (*pScreen->HandleExposures)(pWin);
-#ifdef DO_SAVE_UNDERS
+
+#if XORG < 110 && defined(DO_SAVE_UNDERS)
 	if (dosave)
 	    (*pScreen->PostChangeSaveUnder)(pLayerWin, pLayerWin);
 #endif /* DO_SAVE_UNDERS */
@@ -1089,11 +1157,14 @@ static Bool
 vfbScreenInit(int index, ScreenPtr pScreen, int argc, char **argv)
 {
     vfbScreenInfoPtr pvfb = &vfbScreens[index];
-    int dpi = 100;
+    int dpi;
     int ret;
     void *pbits;
 
-    if (monitorResolution) dpi = monitorResolution;
+    /* 96 is the default used by most other systems */
+    dpi = 96;
+    if (monitorResolution)
+        dpi = monitorResolution;
 
     pbits = vfbAllocateFramebufferMemory(&pvfb->fb);
     if (!pbits) return FALSE;
@@ -1145,7 +1216,9 @@ vfbScreenInit(int index, ScreenPtr pScreen, int argc, char **argv)
 
     if (!ret) return FALSE;
 
+#if XORG < 110
     miInitializeBackingStore(pScreen);
+#endif
 
     /*
      * Circumvent the backing store that was just initialised.  This amounts
@@ -1164,13 +1237,23 @@ vfbScreenInit(int index, ScreenPtr pScreen, int argc, char **argv)
     pScreen->blackPixel = pvfb->blackPixel;
     pScreen->whitePixel = pvfb->whitePixel;
 
-    if (!pvfb->pixelFormatDefined && pvfb->fb.depth == 16) {
-	pvfb->pixelFormatDefined = TRUE;
-	pvfb->rgbNotBgr = TRUE;
-	pvfb->blueBits = pvfb->redBits = 5;
-	pvfb->greenBits = 6;
+    if (!pvfb->pixelFormatDefined) {
+	switch (pvfb->fb.depth) {
+	case 16:
+	    pvfb->pixelFormatDefined = TRUE;
+	    pvfb->rgbNotBgr = TRUE;
+	    pvfb->blueBits = pvfb->redBits = 5;
+	    pvfb->greenBits = 6;
+	    break;
+	case 24:
+	case 32:
+	    pvfb->pixelFormatDefined = TRUE;
+	    pvfb->rgbNotBgr = TRUE;
+	    pvfb->blueBits = pvfb->redBits = pvfb->greenBits = 8;
+	    break;
+	}
     }
-    
+
     if (pvfb->pixelFormatDefined) {
 	VisualPtr vis = pScreen->visuals;
 	for (int i = 0; i < pScreen->numVisuals; i++) {
@@ -1303,7 +1386,7 @@ Bool LegalModifier(unsigned int key, DeviceIntPtr pDev)
 void ProcessInputEvents()
 {
   mieqProcessInputEvents();
-#ifdef XORG_15
+#if XORG == 15
   miPointerUpdate();
 #endif
 }
@@ -1315,3 +1398,9 @@ void InitInput(int argc, char *argv[])
 {
   mieqInit ();
 }
+
+#if XORG > 17
+void CloseInput(void)
+{
+}
+#endif
